## NOTE ##
# Some of these tests spawn an underlying server, it might be better not run
# unattended. Some also require a miner, we might not have it handy.
import asyncio

import pytest
import luxos.asyncops
from luxos import utils


@pytest.mark.manual
@pytest.mark.asyncio
async def test_basic_call(echopool):
    """places few round trip calls"""
    echopool.start(1)
    host, port = echopool.addresses[0]

    # this will bomb at around 10k
    for index in range(1_000):
        ret = await luxos.asyncops.roundtrip(
            host, port, f"hello world {index}", asjson=False
        )
        assert ret == f"received by ('{host}', {port}): hello world {index}"


@pytest.mark.manual
@pytest.mark.asyncio
async def test_basic_batched_call(echopool):
    """places few round trip calls"""
    echopool.start(1)
    host, port = echopool.addresses[0]

    # this will bomb at around 10k if run sequentially!
    calls = [
        luxos.asyncops.roundtrip(host, port, f"hello world {index}", asjson=False)
        for index in range(10_000)
    ]
    expected = {
        f"received by ('{host}', {port}): hello world {index}"
        for index in range(len(calls))
    }
    found = set()
    from luxos.misc import batched

    for group in batched(calls, 100):
        tasks = await asyncio.gather(*group, return_exceptions=True)
        found.update(tasks)
    assert expected == found


@pytest.mark.manual
@pytest.mark.asyncio
async def test_util_launch_oneshot(echopool):
    """places few round trip calls"""
    echopool.start(1)
    host, port = echopool.addresses[0]

    async def call(index, *_):
        return await luxos.asyncops.roundtrip(
            host, port, f"hello world {index}", asjson=False
        )

    addresses = [(index, None) for index in range(10)]

    expected = {
        f"received by ('{host}', {port}): hello world {index}" for index, _ in addresses
    }

    ret = await utils.launch(addresses, call)
    assert set(ret) == expected


@pytest.mark.manual
@pytest.mark.asyncio
async def test_util_launch_batched(echopool):
    """places *many* round trip calls"""

    echopool.start(1)
    host, port = echopool.addresses[0]

    async def call(index, *_):
        return await luxos.asyncops.roundtrip(
            host, port, f"hello world {index}", asjson=False
        )

    # utils.launch will choke at 10k without batch argument
    addresses = [(index, None) for index in range(10_000)]
    expected = {
        f"received by ('{host}', {port}): hello world {index}" for index, _ in addresses
    }

    ret = await utils.launch(addresses, call, batch=100)
    assert set(ret) == expected



# catch-all module (find later a better place)
from __future__ import annotations
import sys
import itertools
import ipaddress
from typing import Generator


if sys.version_info >= (3, 12):
    batched = itertools.batched
else:

    def batched(iterable, n):
        if n < 1:
            raise ValueError("n must be at least one")
        it = iter(iterable)
        while batch := tuple(itertools.islice(it, n)):
            yield batch


def indent(txt: str, pre: str = " " * 2) -> str:
    """simple text indentation"""

    from textwrap import dedent

    txt = dedent(txt)
    if txt.endswith("\n"):
        last_eol = "\n"
        txt = txt[:-1]
    else:
        last_eol = ""

    result = pre + txt.replace("\n", "\n" + pre) + last_eol
    return result if result.strip() else result.strip()


def iter_ip_ranges(txt: str) -> Generator[str, None, None]:
    """iterate over ip ranges

    for ip in iter_ip_ranges("127.0.0.1-127.0.0.3:127.0.0.15"):
        print(ip)

    127.0.0.1
    127.0.0.2
    127.0.0.3
    127.0.0.15
    """

    for segment in txt.split(":"):
        start, _, end = segment.partition("-")
        if not end:
            yield start
        else:
            cur = ipaddress.IPv4Address(start)
            last = ipaddress.IPv4Address(end)
            while cur <= last:
                yield str(cur)
                cur += 1



from __future__ import annotations

import functools
import logging
import asyncio
import json

from typing import Any

from . import exceptions
from . import api


log = logging.getLogger(__name__)

TIMEOUT = 3.0  # default timeout for operations
RETRY = 0  # default number (>1) of retry on a failed operation
RETRY_DELAY = 1.0  # delay between retry


def wrapped(function):
    """wraps a function acting on a host and re-raise with internal exceptions

    This re-raise exceptions so they all derive from MinerConnectionError, eg:
    @wrapped
    def somcode(host: str, port: int, ...):
        ...

    try:
        await somecode()
    except MinerConnectionError as e:  <- this will catch all exceptions!
        e.address
        raise MyNewExecption() from e  <- this will re-raise
    """

    @functools.wraps(function)
    async def _function(host: str, port: int, *args, **kwargs):
        try:
            return await function(host, port, *args, **kwargs)
        except asyncio.TimeoutError as e:
            # we augment underlying TimeOuts
            raise exceptions.MinerCommandTimeoutError(host, port) from e
        except exceptions.MinerConnectionError:
            raise
        except Exception as e:
            # we augment any other exception with (host, port) info
            log.exception("internal error")
            raise exceptions.MinerConnectionError(host, port, "internal error") from e

    return _function


async def _roundtrip(
    host: str, port: int, cmd: bytes | str, timeout: float | None
) -> str:
    """simple asyncio socket based send/receive function

    Example:
        print(await _roundtrip(host, port, version))
        -> (str) "{'STATUS': [{'Code': 22, 'Description'...."
    """
    reader, writer = await asyncio.wait_for(
        asyncio.open_connection(host, port), timeout
    )

    writer.write(cmd.encode() if isinstance(cmd, str) else cmd)
    await writer.drain()

    response = bytearray()
    while True:
        data = await asyncio.wait_for(reader.read(1), timeout=timeout)
        if not data:
            break
        null_index = data.find(b"\x00")
        if null_index >= 0:
            response += data[:null_index]
            break
        response += data

    return response.decode()


# TODO add annotations
async def roundtrip(
    host: str,
    port: int,
    cmd: bytes | str | dict[str, Any],
    asjson: bool | None = True,
    timeout: float | None = None,
    retry: int | None = 0,
    retry_delay: float | None = None,
):
    """utility wrapper around _roundrip

    Example:
        print(await roundtrip(host, port, {"version"}))
        -> (json) {'STATUS': [{'Code': 22, 'Description': 'LUXminer 20 ...
        print(await roundtrip(host, port, "version"))
        -> (str) "{'STATUS': [{'Code': 22, 'Description': 'LUXminer 20 ..
    """
    timeout = TIMEOUT if timeout is None else timeout
    retry = RETRY if retry is None else retry
    retry_delay = RETRY_DELAY if retry_delay is None else retry_delay

    if not isinstance(cmd, (bytes, str)):
        cmd = json.dumps(cmd, indent=2, sort_keys=True)
        if asjson is None:
            asjson = True

    last_exception = None
    for _ in range(max(retry, 1)):
        try:
            res = await _roundtrip(host, port, cmd, timeout)
            if asjson:
                return json.loads(res)
            else:
                return res
        except (Exception, asyncio.TimeoutError) as e:
            last_exception = e
        if retry_delay:
            await asyncio.sleep(retry_delay)

    if last_exception is not None:
        raise last_exception


def validate_message(
    host: str,
    port: int,
    res: dict[str, Any],
    extrakey: str | None = None,
    minfields: None | int = None,
    maxfields: None | int = None,
) -> dict[str, Any]:
    # all miner message comes with a STATUS
    for key in ["STATUS", "id", *([extrakey] if extrakey else [])]:
        if key in res:
            continue
        raise exceptions.MinerCommandMalformedMessageError(
            host, port, f"missing {key} from logon message", res
        )

    if not extrakey or not (minfields or maxfields):
        return res

    n = len(res[extrakey])
    msg = None
    if minfields and (n < minfields):
        msg = f"found {n} fields for {extrakey} invalid: " f"{n} <= {minfields}"
    elif maxfields and (n > maxfields):
        msg = f"found {n} fields for {extrakey} invalid: " f"{n} >= {maxfields}"
    if msg is None:
        return res[extrakey]
    raise exceptions.MinerCommandMalformedMessageError(host, port, msg, res)


@wrapped
async def logon(host: str, port: int, timeout: float | None = 3) -> str:
    timeout = TIMEOUT if timeout is None else timeout
    res = await roundtrip(host, port, {"command": "logon"}, timeout=timeout)

    # when we first logon, we'll receive a token (session_id)
    #   [STATUS][SessionID]
    # on subsequent logon, we receive a
    #   [STATUS][Msg] == "Another session is active" ([STATUS][Code] 402)
    if "SESSION" not in res and res.get("STATUS", [{}])[0].get("Code") == 402:
        raise exceptions.MinerCommandSessionAlreadyActive(
            host, port, "connection active", res
        )
    sessions = validate_message(host, port, res, "SESSION", 1, 1)

    session = sessions[0]  # type: ignore

    if "SessionID" not in session:
        raise exceptions.MinerCommandSessionAlreadyActive(
            host, port, "no SessionID in data", res
        )
    return str(session["SessionID"])


@wrapped
async def logoff(
    host: str, port: int, sid: str, timeout: float | None = 3
) -> dict[str, Any]:
    timeout = TIMEOUT if timeout is None else timeout
    return await roundtrip(
        host, port, {"command": "logoff", "parameter": sid}, timeout=timeout
    )


@wrapped
async def execute_command(
    host: str,
    port: int,
    timeout_sec: float | None,
    cmd: str,
    parameters: list[str] | None = None,
    verbose: bool = False,
    asjson: bool | None = True,
    add_address: bool = False,
) -> tuple[tuple[str, int], dict[str, Any]] | dict[str, Any]:
    timeout = TIMEOUT if timeout_sec is None else timeout_sec
    parameters = parameters or []

    sid = None
    if api.logon_required(cmd):
        sid = await logon(host, port)
        parameters = [sid, *parameters]
        log.info("session id requested & obtained for %s:%i (%s)", host, port, sid)
    else:
        log.debug("no logon required for command '%s' on %s:%i", cmd, host, port)

    try:
        packet = {"command": cmd}
        if parameters:
            packet["parameter"] = ",".join(parameters)
        log.debug(
            "executing command '%s' on '%s:%i' with parameters: %s",
            cmd,
            host,
            port,
            packet.get("parameter", ""),
        )
        ret = await roundtrip(host, port, packet, timeout=timeout, asjson=asjson)
        log.debug("received from %s:%s: %s", host, port, str(ret))
        return ((host, port), ret) if add_address else ret
    finally:
        if sid:
            await logoff(host, port, sid)


def _rexec_paramteres(
    parameters: str | list[Any] | dict[str, Any] | None = None,
) -> list[str]:
    if isinstance(parameters, dict):
        data = []
        for key, value in parameters.items():
            if value is None:
                value = "null"
            elif value is True:
                value = "true"
            elif value is False:
                value = "false"
            data.append(f"{key}={value}")
        parameters = data
    parameters = ([parameters] if isinstance(parameters, str) else parameters) or []
    parameters = [str(param) for param in parameters]
    return parameters


async def rexec(
    host: str,
    port: int,
    cmd: str,
    parameters: str | list[Any] | dict[str, Any] | None = None,
    timeout: float | None = None,
    retry: int | None = None,
    retry_delay: float | None = None,
) -> dict[str, Any] | None:
    from . import api

    parameters = _rexec_paramteres(parameters)

    timeout = TIMEOUT if timeout is None else timeout
    retry = RETRY if retry is None else retry
    retry_delay = RETRY_DELAY if retry_delay is None else retry_delay

    # if cmd is logon/logoff we dealt with it differently
    if cmd in {"logon", "logoff"}:
        failure = None
        for i in range(retry or 1):
            try:
                if cmd == "logon":
                    return {"sid": await logon(host, port, timeout)}
                else:
                    return await logoff(host, port, parameters[0])
            except Exception as exc:
                failure = exc
            if retry and (i < retry) and retry_delay:
                await asyncio.sleep(retry_delay)
        if isinstance(failure, Exception):
            raise failure

    failure = None
    sid = ""
    for i in range(retry + 1):
        if not api.logon_required(cmd):
            log.debug("no logon required for command '%s' on %s:%i", cmd, host, port)
            break
        try:
            sid = await logon(host, port, timeout)
            parameters = [sid, *parameters]
            log.info("session id requested & obtained for %s:%i (%s)", host, port, sid)
            break
        except Exception as exc:
            failure = exc
        if retry and (i < retry) and retry_delay:
            await asyncio.sleep(retry_delay)

    if isinstance(failure, Exception):
        raise failure

    packet = {"command": cmd}
    if parameters:
        packet["parameter"] = ",".join(parameters)
    log.debug(
        "executing command '%s' on '%s:%i' with parameters: %s",
        cmd,
        host,
        port,
        packet.get("parameter", ""),
    )

    failure = None
    for i in range(retry + 1):
        try:
            ret = await roundtrip(host, port, packet, timeout=timeout)
            log.debug("received from %s:%s: %s", host, port, str(ret))
            if sid:
                await logoff(host, port, sid)
            return ret
        except Exception as exc:
            failure = exc
        if retry and (i < retry) and retry_delay:
            log.debug("failed attempt %i (out of %i)", i + 1, retry)
            await asyncio.sleep(retry_delay)

    if sid:
        await logoff(host, port, sid)
    if isinstance(failure, Exception):
        raise failure



"""cli utilities

This is the v1.cli decorator to help writing cli scripts with a consistent
interface.

A simple v1 script will always have:
* `-v/--verbose | -q/--quiet` flags to increase the logging verbosity level
* `-c/--config` to pass a config file (default to config.yaml)

A `sample.py` script with default sensible and consistent interface:

```
import argparse
import luxos.cli.v1 as cli
import logging

log = logging.getLogger(__name__)


@cli.cli()
def main(args: argparse.Namespace):
    log.debug("a debug message, need to use -v|--verbose to display it")
    log.info("an info message, you can silence it with -q|--quiet")
    log.warning("a warning!")


if __name__ == "__main__":
    main()
```


### Advanced usages

#### changing the default config file

In the `sample.py` file, just add:

```
CONFIGPATH = "my.config.file.yaml"
```

`CONFIGPATH` is part of the module level "magic" variables.

Another one is `LOGGING_CONFIG` (not recommended!):
```
LOGGING_CONFIG = {
    'level': logging.INFO,
    'format': "%(asctime)s:%(name)s:[%(levelname)s] %(message)s",
    'handlers': [
        logging.StreamHandler(),
        logging.FileHandler("LuxOS-LoadControl.log")
    ],
}
```

The `LOGGING_CONFIG` is a dictionary feed into `logging.basicConfig(**LOGGING_CONFIG)`.

#### add and process new extra arguments

in the `sample.py` file:

```
def add_arguments(parser: argparse.ArgumentParser) -> None:
    ... adds as many args needed


def process_args(args: argparse.Namespace) -> argparse.Namespace | None:
    ... you can manipulate args in place, changing its attributes in place
    ... if you return a new Namespace instance then it will be fee to main below


@cli.cli(add_arguments, process_args)
def main(args: argparse.Namespace):
    ... args is the Namespace instance returned by process_args if defined,
    ... or the default processed one of process_args is not provided

```

#### escape hatch to handle all the parsing/processing

```

@cli.cli()
def main(parser: argparse.ArgumentParser):
    ... you can use directly the parser here

```

"""

from __future__ import annotations

import sys
import contextlib
import inspect
import logging
import time
from pathlib import Path
import functools
import argparse
from typing import Callable, Any


# SPECIAL MODULE LEVEL VARIABLES
MODULE_VARIABLES = {
    "LOGGING_CONFIG": None,  # logging config dict (passed to logging.basicConfig(**LOGGING_CONFIG))
    "CONFIGPATH": Path("config.yaml"),  # config default path
}


log = logging.getLogger(__name__)


class CliBaseError(Exception):
    pass


class AbortCliError(CliBaseError):
    pass


class AbortWrongArgument(CliBaseError):
    pass


def setup_logging(config: dict[str, Any], count: int) -> None:
    levelmap = [
        logging.WARNING,
        logging.INFO,
        logging.DEBUG,
    ]
    n = len(levelmap)

    # awlays start from info level
    level = logging.INFO

    # we can set the default start log level in LOGGING_CONFIG
    if config.get("level", None) is not None:
        level = config["level"]

    # we control if we go verbose or quite here
    index = levelmap.index(level) + count
    config["level"] = levelmap[max(min(index, n - 1), 0)]
    logging.basicConfig(**config)


class LuxosParser(argparse.ArgumentParser):
    def __init__(self, module_variables, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.module_variables = module_variables or {}

        # we're adding the -v|-q flags, to control the logging level
        self.add_argument(
            "-v", "--verbose", action="count", help="report verbose logging"
        )
        self.add_argument("-q", "--quiet", action="count", help="report quiet logging")

        # we add the -c|--config flag to point to a config file
        configpath = (
            self.module_variables.get("CONFIGPATH") or MODULE_VARIABLES["CONFIGPATH"]
        )
        configpath = Path(configpath).expanduser().absolute()
        if configpath.is_relative_to(Path.cwd()):
            configpath = configpath.relative_to(Path.cwd())

        self.add_argument(
            "-c",
            "--config",
            default=configpath,
            type=Path,
            help="path to a config file",
        )

    def parse_args(self, args=None, namespace=None):
        options = super().parse_args(args, namespace)

        # we provide an error function to nicely bail out the script
        def error(msg):
            raise AbortWrongArgument(msg)

        self.error = error
        options.error = self.error

        # setup the logging
        config = {}
        if value := self.module_variables.get("LOGGING_CONFIG"):
            config = value.copy()

        count = (options.verbose or 0) - (options.quiet or 0)
        setup_logging(config, count)

        return options

    @classmethod
    def get_parser(cls, module_variables, **kwargs):
        class Formatter(
            argparse.ArgumentDefaultsHelpFormatter, argparse.RawTextHelpFormatter
        ):
            pass

        return cls(
            module_variables=module_variables, formatter_class=Formatter, **kwargs
        )


def cli(
    add_arguments: Callable[[argparse.ArgumentParser], None] | None = None,
    process_args: (
        Callable[[argparse.Namespace], argparse.Namespace | None] | None
    ) = None,
):
    def _cli1(function):
        @contextlib.contextmanager
        def setup():
            sig = inspect.signature(function)

            module_variables = MODULE_VARIABLES.copy()
            module = inspect.getmodule(function)
            for name in list(module_variables):
                module_variables[name] = getattr(module, name, None)

            if "args" in sig.parameters and "parser" in sig.parameters:
                raise RuntimeError("cannot use args and parser at the same time")

            description, _, epilog = (
                (function.__doc__ or module.__doc__ or "").strip().partition("\n")
            )
            kwargs = {}
            parser = LuxosParser.get_parser(
                module_variables, description=description, epilog=epilog
            )
            if add_arguments:
                add_arguments(parser)

            if "parser" in sig.parameters:
                kwargs["parser"] = parser

            t0 = time.monotonic()
            success = "completed"
            errormsg = ""
            show_timing = True
            try:
                if "parser" not in sig.parameters:
                    args = parser.parse_args()
                    if process_args:
                        args = process_args(args) or args
                    if "args" in sig.parameters:
                        kwargs["args"] = args
                yield sig.bind(**kwargs)
            except AbortCliError as exc:
                errormsg = str(exc)
                success = "failed"
            except AbortWrongArgument as exc:
                show_timing = False
                parser.print_usage(sys.stderr)
                print(f"{parser.prog}: error: {exc.args[0]}", file=sys.stderr)
                sys.exit(2)
            except Exception:
                log.exception("un-handled exception")
                success = "failed"
            finally:
                if show_timing:
                    delta = round(time.monotonic() - t0, 2)
                    log.info("task %s in %.2fs", success, delta)
            if errormsg:
                parser.error(errormsg)

        def log_sys_info():
            log.debug("interpreter: %s", sys.executable)
            log.debug("version: %s", sys.version)

        if inspect.iscoroutinefunction(function):

            @functools.wraps(function)
            async def _cli2(*args, **kwargs):
                with setup() as ba:
                    log_sys_info()
                    return await function(*ba.args, **ba.kwargs)

        else:

            @functools.wraps(function)
            def _cli2(*args, **kwargs):
                with setup() as ba:
                    log_sys_info()
                    return function(*ba.args, **ba.kwargs)

        return _cli2

    return _cli1



from __future__ import annotations
import asyncio
import os
import csv
import json
import time
import socket
import logging
import argparse
import ipaddress
import threading
from typing import Any

from luxos.api import logon_required

log = logging.getLogger(__name__)


# internal_send_cgminer_command sends a command to the cgminer API server and returns the response.
def internal_send_cgminer_command(
    host: str, port: int, command: str, timeout_sec: int, verbose: bool
) -> dict[str, Any]:
    # Create a socket connection to the server
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        try:
            # set timeout
            sock.settimeout(timeout_sec)

            # Connect to the server
            sock.connect((host, port))

            # Send the command to the server
            sock.sendall(command.encode())

            # Receive the response from the server
            response = bytearray()
            while True:
                # Read one byte at a time so we can wait for the null terminator.
                # this is to avoid waiting for the timeout as we don't know how long
                # the response will be and socket.recv() will block until reading
                # the specified number of bytes.
                try:
                    data = sock.recv(1)
                except socket.timeout:
                    # Timeout occurred, check if we have any data so far
                    if len(response) == 0:
                        raise ValueError("timeout waiting for data")
                    else:
                        break
                if not data:
                    break
                null_index = data.find(b"\x00")
                if null_index >= 0:
                    response += data[:null_index]
                    break
                response += data

            # Parse the response JSON
            r = json.loads(response.decode())
            log.debug(r)
            return r

        except socket.error as e:
            raise e


# send_cgminer_command sends a command to the cgminer API server and returns the response.
def send_cgminer_command(
    host: str, port: int, cmd: str, param: str, timeout: int, verbose: bool
) -> dict[str, Any]:
    req = str(f'{{"command": "{cmd}", "parameter": "{param}"}}\n')
    log.debug(f"Executing command: {cmd} with params: {param} to host: {host}")

    return internal_send_cgminer_command(host, port, req, timeout, verbose)


# send_cgminer_simple_command sends a command with no params to the miner and returns the response.
def send_cgminer_simple_command(
    host: str, port: int, cmd: str, timeout: int, verbose: bool
) -> dict[str, Any]:
    req = str(f'{{"command": "{cmd}"}}\n')
    log.debug(f"Executing command: {cmd} to host: {host}")
    return internal_send_cgminer_command(host, port, req, timeout, verbose)


# check_res_structure checks that the response has the expected structure.
def check_res_structure(
    res: dict[str, Any], structure: str, min: int, max: int
) -> dict[str, Any]:
    # Check the structure of the response.
    if structure not in res or "STATUS" not in res or "id" not in res:
        raise ValueError("error: invalid response structure")

    # Should we check min and max?
    if min >= 0 and max >= 0:
        # Check the number of structure elements.
        if not (min <= len(res[structure]) <= max):
            raise ValueError(
                f"error: unexpected number of {structure} in response; min: {min}, max: {max}, actual: {len(res[structure])}"
            )

    # Should we check only min?
    if min >= 0:
        # Check the minimum number of structure elements.
        if len(res[structure]) < min:
            raise ValueError(
                f"error: unexpected number of {structure} in response; min: {min}, max: {max}, actual: {len(res[structure])}"
            )

    # Should we check only max?
    if max >= 0:
        # Check the maximum number of structure elements.
        if len(res[structure]) < min:
            raise ValueError(
                f"error: unexpected number of {structure} in response; min: {min}, max: {max}, actual: {len(res[structure])}"
            )

    return res


# get_str_field tries to get the field as a string and returns it.
def get_str_field(struct: dict[str, Any], name: str) -> str:
    try:
        s = str(struct[name])
    except Exception as e:
        raise ValueError(f"error: invalid {name} str field: {e}")

    return s


def logon(host: str, port: int, timeout: int, verbose: bool) -> str:
    # Send 'logon' command to cgminer and get the response
    res = send_cgminer_simple_command(host, port, "logon", timeout, verbose)

    # Check if the response has the expected structure
    check_res_structure(res, "SESSION", 1, 1)

    # Extract the session data from the response
    session = res["SESSION"][0]

    # Get the 'SessionID' field from the session data
    s = get_str_field(session, "SessionID")

    # If 'SessionID' is empty, raise an error indicating invalid session id
    if s == "":
        raise ValueError("error: invalid session id")

    # Return the extracted 'SessionID'
    return s


def add_session_id_parameter(session_id, parameters):
    # Add the session id to the parameters
    return [session_id, *parameters]


def parameters_to_string(parameters):
    # Convert the parameters to a string that LuxOS API accepts
    return ",".join(parameters)


def generate_ip_range(start_ip: str, end_ip: str) -> list[str]:
    # Generate a list of IP addresses from the start and end IP
    start = ipaddress.IPv4Address(start_ip)
    end = ipaddress.IPv4Address(end_ip)

    ip_list = []

    while start <= end:
        ip_list.append(str(start))
        start += 1

    return ip_list


def load_ip_list_from_csv(csv_file: str) -> list[str]:
    # Check if file exists
    if not os.path.exists(csv_file):
        logging.info(f"Error: {csv_file} file not found.")
        exit(1)

    # Load the IP addresses from the CSV file
    ip_list = []
    with open(csv_file, "r") as f:
        reader = csv.reader(f)
        for i, row in enumerate(reader):
            # skip commented lines
            if row and row[0].strip().startswith("#"):
                continue
            if i == 0 and row and row[0] == "hostname":
                continue
            if row:  # Ignore empty rows
                ip_list.extend(row)
    return ip_list


def execute_command(
    host: str, port: int, timeout_sec: int, cmd: str, parameters: list, verbose: bool
):
    # Check if logon is required for the command
    logon_req = logon_required(cmd)

    try:
        if logon_req:
            # Get a SessionID
            sid = logon(host, port, timeout_sec, verbose)
            # Add the SessionID to the parameters list at the left.
            parameters = add_session_id_parameter(sid, parameters)

            log.debug("Command requires a SessionID, logging in for host: %s", host)
            log.info("SessionID obtained for %s: %s", host, sid)

        # TODO verify this
        elif not logon_required:  # type: ignore
            log.debug("Logon not required for executing %s", cmd)

        # convert the params to a string that LuxOS API accepts
        param_string = parameters_to_string(parameters)

        log.debug("%s on %s with parameters: %s", cmd, host, param_string)

        # Execute the API command
        res = send_cgminer_command(host, port, cmd, param_string, timeout_sec, verbose)

        log.debug(res)

        # Log off to terminate the session
        if logon_req:
            send_cgminer_command(host, port, "logoff", sid, timeout_sec, verbose)

        return res

    except Exception:
        log.exception("Error executing %s on %s", cmd, host)


def main():
    # define arguments
    parser = argparse.ArgumentParser(description="LuxOS CLI Tool")
    parser.add_argument("--range_start", required=False, help="IP start range")
    parser.add_argument("--range_end", required=False, help="IP end range")
    parser.add_argument(
        "--ipfile",
        required=False,
        default="ips.csv",
        help="File name to store IP addresses",
    )
    parser.add_argument("--cmd", required=True, help="Command to execute on LuxOS API")
    parser.add_argument(
        "--params",
        required=False,
        default=[],
        nargs="+",
        help="Parameters for LuxOS API",
    )
    parser.add_argument(
        "--max_threads",
        required=False,
        default=10,
        type=int,
        help="Maximum number of threads to use. Default is 10.",
    )
    parser.add_argument(
        "--timeout",
        required=False,
        default=3,
        type=int,
        help="Timeout for network scan in seconds",
    )
    parser.add_argument(
        "--port", required=False, default=4028, type=int, help="Port for LuxOS API"
    )
    parser.add_argument("--verbose", action="store_true", help="Verbose output")
    parser.add_argument(
        "--batch_delay",
        required=False,
        default=0,
        type=int,
        help="Delay between batches in seconds",
    )
    parser.add_argument(
        "--async",
        dest="async_engine",
        action="store_true",
        help="enable the new async engine",
    )
    parser.add_argument(
        "-a",
        "--all",
        dest="details",
        action="store_true",
        help="show full result output",
    )

    # parse arguments
    args = parser.parse_args()
    args.error = parser.error

    logging.basicConfig(
        level=logging.DEBUG if args.verbose else logging.INFO,
        format="%(asctime)s [%(levelname)s] %(message)s",
        handlers=[logging.StreamHandler(), logging.FileHandler("LuxOS-CLI.log")],
    )

    # set timeout to milliseconds
    timeout_sec = args.timeout

    # check if IP address range or CSV with list of IP is provided
    if args.range_start and args.range_end:
        ip_list = generate_ip_range(args.range_start, args.range_end)
    elif args.ipfile:
        ip_list = load_ip_list_from_csv(args.ipfile)
    else:
        args.error("No IP address or IP list found.")

    # Set max threads to use, minimum of max threads and number of IP addresses
    max_threads = min(args.max_threads, len(ip_list))

    # Set start time
    start_time = time.monotonic()

    if args.async_engine:
        from . import async_luxos

        asyncio.run(
            async_luxos.run(
                ipaddresses=ip_list,
                port=args.port,
                cmd=args.cmd,
                params=args.params,
                timeout=timeout_sec,
                delay=args.batch_delay,
                details=args.details,
                batchsize=args.max_threads,
            )
        )

        # TODO remove this duplicate code
        end_time = time.monotonic()
        execution_time = end_time - start_time
        log.info(f"Execution completed in {execution_time:.2f} seconds.")
        return

    # Create a list of threads
    threads = []

    # Iterate over the IP addresses
    for ip in ip_list:
        # create new thread for each IP address
        thread = threading.Thread(
            target=execute_command,
            args=(ip, args.port, timeout_sec, args.cmd, args.params, args.verbose),
        )

        # start the thread
        threads.append(thread)
        thread.start()

        # Limit the number of concurrent threads
        if len(threads) >= max_threads:
            # Wait for the threads to finish
            for thread in threads:
                thread.join()

            # Introduce the batch delay if specified
            if args.batch_delay > 0:
                print(f"Waiting {args.batch_delay} seconds")
                time.sleep(args.batch_delay)

            # Clear the thread list for the next batch
            threads = []

    # Wait for the remaining threads to finish
    for thread in threads:
        thread.join()

    # Execution completed
    end_time = time.monotonic()
    execution_time = end_time - start_time
    log.info(f"Execution completed in {execution_time:.2f} seconds.")


if __name__ == "__main__":
    main()



class LuxosBaseException(Exception):
    pass


class MinerConnectionError(LuxosBaseException):
    def __init__(self, host: str, port: int, *args, **kwargs):
        super().__init__(host, port, *args, **kwargs)
        self.address = (host, port)

    def __str__(self):
        return (f"<{self.address[0]}:{self.address[1]}>: {self.__class__.__name__}, "
                f"{self.args[2] if self.args[2:] else 'unknown reason'}")


class MinerCommandSessionAlreadyActive(MinerConnectionError):
    pass


class MinerCommandTimeoutError(MinerConnectionError):
    pass


class MinerCommandMalformedMessageError(MinerConnectionError):
    pass



from luxos import misc


def test_misc_batched():
    data = {}
    for index, group in enumerate(misc.batched(range(10), 3)):
        data[index] = group

    assert data == {
        0: (0, 1, 2),
        1: (3, 4, 5),
        2: (6, 7, 8),
        3: (9,),
    }


def test_indent():
    txt = """\
     An unusually complicated text
    with un-even indented lines
   that make life harder
"""
    assert (
        misc.indent(txt, pre="..")
        == """\
..  An unusually complicated text
.. with un-even indented lines
..that make life harder
"""
    )


def test_iter_ip_ranges():
    assert set(misc.iter_ip_ranges("127.0.0.1")) == {"127.0.0.1"}
    assert set(misc.iter_ip_ranges("127.0.0.1-127.0.0.3")) == {
        "127.0.0.1",
        "127.0.0.2",
        "127.0.0.3",
    }
    assert set(misc.iter_ip_ranges("127.0.0.1-127.0.0.3:127.0.0.15")) == {
        "127.0.0.1",
        "127.0.0.2",
        "127.0.0.3",
        "127.0.0.15",
    }
